from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING
from damask import __version__
from damask_parse.writers import write_load_case

if TYPE_CHECKING:
    from matflow.param_classes.load import LoadCase


def write_load(
    path: Path | str, load_case: LoadCase, damask_solver: dict[str, str] | None
) -> None:
    """Write the load file for a DAMASK simulation.

    Parameters
    ----------
    path
        Full path to load file to write.
    load_case
        The loadings to write.
    solver
        Override information about the solver to use.
    """
    path_ = Path(path)
    write_load_case(
        dir_path=path_.parent,
        load_cases=load_case.create_damask_loading_plan(),
        solver=damask_solver,
        name=path_.name,
        write_2D_arrs=(__version__ != "3.0.0-alpha3"),
    )



import numpy as np


def estimate_cov(indicator, p_i: float) -> float:
    """Estimate the coefficient of variation at a given conditional level of the subset
    simulation."""

    num_chains, num_states = indicator.shape
    N = num_chains * num_states  # samples per level

    # covariance sequence (estimated), Eq. 29
    r = np.zeros(num_states - 1)
    for k in range(num_states - 1):
        r_k = 0
        for l in range(num_states - (k + 1)):
            i_1 = indicator[:, l]
            i_2 = indicator[:, l + k + 1]
            r_k_i = np.dot(i_1, i_2)
            r_k += r_k_i
        r[k] = (r_k / (N - (k + 1) * num_chains)) - p_i**2

    r_0 = np.sum(indicator**2) / N - p_i**2  # autocovariance at lag zero (exact)
    r_0 = p_i * (1 - p_i)

    rho = r / r_0

    gamma = 2 * sum(
        (1 - (k + 1) * num_chains / N) * rho[k] for k in range(num_states - 1)
    )

    delta = np.sqrt((1 - p_i) / (p_i * N) * (1 + gamma))

    return delta


def collate_results(g, x, p_0, all_g, all_x, all_accept):

    # all iterations of g are passed just to get the level index:
    # TODO: in future set and read loop_idx from environment variable?

    level_idx = sorted(g.items(), key=lambda x: int(x[0].split("_")[1]))[-1][1][
        "loop_idx"
    ]["levels"]

    g = g["iteration_0"]["value"]
    num_samples = len(g)

    if all_g:
        # from multiple Markov chains:
        g = np.concatenate([i[:] for i in all_g])
        accept = np.vstack([i[:] for i in all_accept])
        x = np.vstack([i[:] for i in all_x])
        accept_rate = np.mean(accept)
    else:
        # from initial direct Monte Carlo samples:
        g = np.array(g)
        x = np.vstack([i[:] for i in x])
        accept_rate = None

        # consider missing data in `g` by copying from non-missing data:
        bad_bool = g == None
        if np.any(bad_bool):
            bad_idx = np.where(bad_bool)[0]
            print(
                f"Collate results: missing {len(bad_idx)} system performance "
                f"evaluation(s) at indices: {bad_idx!r}."
            )
            good_idx = np.where(~bad_bool)[0]
            replace_idx = good_idx[: len(bad_idx)]
            replace_g = g[replace_idx]
            replace_x = x[replace_idx]
            x[bad_idx] = replace_x
            g_corrected = np.empty(len(g))
            g_corrected[good_idx] = g[good_idx]
            g_corrected[bad_idx] = replace_g
            g = g_corrected

    num_failed = int(np.sum(g > 0))
    num_chains = int(len(g) * p_0)
    num_states = int(num_samples / num_chains)

    # sort responses
    srt_idx = np.argsort(g)[::-1]  # sort by closest-to-failure first
    g = g[srt_idx]
    x = x[srt_idx, :]

    threshold = (g[num_chains - 1] + g[num_chains]) / 2

    # failure probability at this level:
    fail_bool = g > 0
    level_pf = np.mean(fail_bool) if threshold > 0 else p_0

    chain_seeds = x[:num_chains]
    chain_g = g[:num_chains]

    is_finished = (num_failed / num_samples) >= p_0

    pf = p_0**level_idx * num_failed / num_samples

    if all_g:
        # from multiple Markov chains:
        indicator = np.reshape(
            g > np.minimum(threshold, 0), (num_chains, num_states)
        ).astype(int)
        level_cov = estimate_cov(indicator, level_pf)
    else:
        # from initial direct Monte Carlo samples:
        level_cov = np.sqrt((1 - level_pf) / (num_samples * level_pf))

    return {
        "chain_seeds": chain_seeds,
        "chain_g": chain_g,
        "threshold": threshold,
        "num_chains": num_chains,
        "num_failed": num_failed,
        "level_pf": level_pf,
        "level_cov": level_cov,
        "pf": pf,
        "is_finished": is_finished,
        "accept_rate": accept_rate,
    }



"""Functions copied from matflow-defdap package."""
from __future__ import annotations
from pathlib import Path
from typing import Any

import defdap.ebsd as ebsd
import defdap.hrdic as hrdic
from defdap.quat import Quat
import numpy as np
from numpy.typing import NDArray
from scipy.stats import mode
from scipy.ndimage import zoom


def load_microstructure_EBSD_DIC(
    DIC: dict,
    EBSD: dict,
    root_path: str,
    transform_type: str,
    scaling_factor: float,
    find_grains_algorithm: str,
) -> dict:
    ebsd_map = _load_EBSD_map(
        root_path=root_path,
        ebsd_filename=EBSD["filename"],
        ebsd_boundary_tol=EBSD.get("boundary_tol", 10),
        ebsd_min_grain_size=EBSD.get("min_grain_size", 10),
    )
    dic_map = _load_DIC_map(
        root_path=root_path,
        dic_filename=DIC["filename"],
        dic_crop=DIC.get("crop", None),
        dic_scale=DIC.get("scale", None),
    )
    _link_EBSD_DIC_maps(
        ebsd_map=ebsd_map,
        dic_map=dic_map,
        dic_homog_points=DIC["homog_points"],
        dic_min_grain_size=DIC.get("min_grain_size", 10),
        ebsd_homog_points=EBSD["homog_points"],
        transform_type=transform_type,
        find_grains_algorithm=find_grains_algorithm,
    )
    DIC_image = _get_DIC_image(dic_map, scaling_factor)

    return {"microstructure_image": DIC_image}


def _load_EBSD_map(
    root_path: str, ebsd_filename: str, ebsd_boundary_tol: float, ebsd_min_grain_size: int
) -> ebsd.Map:
    "Load EBSD map and detect grains."

    ebsd_map = ebsd.Map(Path(root_path).joinpath(ebsd_filename))

    # check for non-indexed points
    if np.count_nonzero(ebsd_map.phaseArray == 0) != 0:
        raise ValueError("EBSD map contains non-indexed points.")

    ebsd_map.buildQuatArray()

    ebsd_map.findBoundaries(boundDef=ebsd_boundary_tol)
    ebsd_map.findGrains(minGrainSize=ebsd_min_grain_size)
    ebsd_map.calcGrainAvOris()

    return ebsd_map


def _load_DIC_map(
    root_path: str, dic_filename: str, dic_crop: list[int] | None, dic_scale: float | None
) -> hrdic.Map:
    "Load in DIC, crop and set scale."

    dic_map = hrdic.Map(root_path, dic_filename)
    if dic_crop is not None:
        dic_map.setCrop(
            xMin=dic_crop[0], xMax=dic_crop[1], yMin=dic_crop[2], yMax=dic_crop[3]
        )
    if dic_scale is not None:
        dic_map.setScale(dic_scale)

    return dic_map


def _link_EBSD_DIC_maps(
    ebsd_map: ebsd.Map,
    dic_map: hrdic.Map,
    dic_homog_points: NDArray,
    dic_min_grain_size: int,
    ebsd_homog_points: NDArray,
    transform_type: str,
    find_grains_algorithm: str,
) -> None:
    dic_map.homogPoints = dic_homog_points
    ebsd_map.homogPoints = ebsd_homog_points
    dic_map.linkEbsdMap(ebsd_map, transformType=transform_type)

    dic_map.findGrains(minGrainSize=dic_min_grain_size, algorithm=find_grains_algorithm)


def _get_DIC_image(dic_map: hrdic.Map, scaling_factor: float) -> dict[str, Any]:
    # Construct an array of Euler angles
    grain_quats = np.empty((len(dic_map), 4))

    # Transformation orientations from EBSD orientation reference frame
    # to EBSD spatial reference frame
    frame_transform = Quat.fromAxisAngle(np.array((1, 0, 0)), np.pi)

    if dic_map.ebsdMap.crystalSym == "hexagonal":
        # Convert hex convention from y // a2 of EBSD map to x // a1 for DAMASK
        hex_transform = Quat.fromAxisAngle(np.array([0, 0, 1]), -np.pi / 6)
        for i, grain in enumerate(dic_map):
            grain_quats[i] = (
                hex_transform * grain.ebsdGrain.refOri * frame_transform
            ).quatCoef

    else:
        for i, grain in enumerate(dic_map):
            grain_quats[i] = (grain.ebsdGrain.refOri * frame_transform).quatCoef

    # Filter out -1 (grain boundary points) and -2 (too small grains)
    # values in the grain image
    grain_image = dic_map.grains
    _remove_boundary_points(grain_image)
    _remove_small_grain_points(grain_image)
    _remove_boundary_points(grain_image)
    _remove_boundary_points(grain_image, force_remove=True)

    # scale down image if needed
    if scaling_factor != 1:
        grain_image = zoom(
            grain_image, scaling_factor, order=0, prefilter=False, mode="nearest"
        )

    # downstream expects grain numbering to start at 0 not 1
    grain_image -= 1

    # set grain IDs as contiguous range:
    uniq, inv = np.unique(grain_image.reshape(-1, order="C"), return_inverse=True)
    grain_quats = grain_quats[uniq]
    grain_image = np.reshape(inv, grain_image.shape, order="C")

    DIC_image = {
        "orientations": {
            "type": "quat",
            "unit_cell_alignment": {"x": "a"},
            "quaternions": grain_quats,
            "P": 1,  # DefDAP uses P=+1 (e.g see `defdap.quat.Quat.__mul__`)
            "quat_component_ordering": "scalar-vector",
        },
        "grains": grain_image,
    }
    try:
        DIC_image["scale"] = dic_map.scale
    except ValueError:
        pass

    return DIC_image


def _select_area(i: int, j: int, grain_image: NDArray) -> tuple[NDArray, int]:
    i_min, i_max = 1, 1
    j_min, j_max = 1, 1

    on_edge = 0

    if i == 0:
        i_min = 0
        on_edge += 1
    elif i == grain_image.shape[0] - 1:
        i_max = 0
        on_edge += 1

    if j == 0:
        j_min = 0
        on_edge += 1
    elif j == grain_image.shape[1] - 1:
        j_max = 0
        on_edge += 1

    # select 3x3 region around point
    area = grain_image[i - i_min : i + i_max + 1, j - j_min : j + j_max + 1]

    return area, on_edge


def _remove_boundary_points(
    grain_image: NDArray, *, force_remove: bool = False, max_iterations: int = 200
) -> None:
    num_bad_prev = 0
    iteration = 0
    while True:
        num_bad = np.count_nonzero(grain_image == -1)
        if num_bad == 0:
            # No bad values left, done
            print("All bad points removed.")
            break
        elif num_bad == num_bad_prev:
            # Not removing any more
            print("Number of bad points is not decreasing!")
            break
        if iteration > max_iterations:
            print("Max iterations.")
            break

        iteration += 1
        print("Starting iteration {}, num bad: {}".format(iteration, num_bad))

        grain_image_new = np.copy(grain_image)

        # because of how boundaries are defined, best to take from -ve side
        for i, j in zip(*np.where(grain_image == -1)):
            if (
                i != 0
                and j != 0
                and grain_image[i - 1, j] > 0
                and grain_image[i, j - 1] > 0
            ):
                # if both left and above pixels defined, check the diagonal
                if grain_image[i - 1, j - 1] == grain_image[i - 1, j]:
                    grain_image_new[i, j] = grain_image[i - 1, j]

                elif grain_image[i - 1, j - 1] == grain_image[i, j - 1]:
                    grain_image_new[i, j] = grain_image[i, j - 1]

                else:
                    # give up, try in next iteration

                    if force_remove:
                        area, on_edge = _select_area(i, j, grain_image)
                        area = area.flatten()
                        area = area[np.where(area > 0)]  # remove -1 and -2

                        mode_vals, mode_counts = mode(area, keepdims=True)
                        for mode_val, mode_count in zip(mode_vals, mode_counts):
                            #         mode_val, mode_count = mode_vals[0], mode_counts[0]
                            #                         if mode_count >= num_neighbours:
                            grain_image_new[i, j] = mode_val
                            break

            elif i != 0 and grain_image[i - 1, j] > 0:
                grain_image_new[i, j] = grain_image[i - 1, j]

            elif j != 0 and grain_image[i, j - 1] > 0:
                grain_image_new[i, j] = grain_image[i, j - 1]

            # give up, try in next iteration

        num_bad_prev = num_bad
        # [:, :] required to update the array passed in
        grain_image[:, :] = grain_image_new


def _remove_small_grain_points(
    grain_image: NDArray, *, max_iterations: int = 200
) -> None:
    # num_neighbours - must have at least this many pixels surrounding
    # start checking for 8 neighbours, then 7 until 2
    for num_neighbours in list(range(8, 1, -1)):
        print(f"Starting iterations with at least {num_neighbours} equal neighbours")

        num_bad_prev = 0
        iteration = 0
        while True:
            num_bad = np.count_nonzero(grain_image == -2)
            if num_bad == 0:
                # No bad values left, done
                print("All bad points removed.")
                return
            elif num_bad == num_bad_prev:
                # Not removing any more
                print("Number of bad points is not decreasing!")
                break
            if iteration > max_iterations:
                print("Max iterations.")
                break

            iteration += 1
            print("Starting iteration {}, num bad: {}".format(iteration, num_bad))

            grain_image_new = np.copy(grain_image)

            for i, j in zip(*np.where(grain_image == -2)):
                area, on_edge = _select_area(i, j, grain_image)
                area = area.flatten()
                area = area[np.where(area > 0)]  # remove -1 and -2

                mode_vals, mode_counts = mode(area, keepdims=True)
                for mode_val, mode_count in zip(mode_vals, mode_counts):
                    if mode_count >= num_neighbours:
                        grain_image_new[i, j] = mode_val
                        break

            num_bad_prev = num_bad
            # [:, :] required to update the array passed in
            grain_image[:, :] = grain_image_new



from __future__ import annotations
from damask_parse.utils import add_volume_element_buffer_zones


def add_VE_buffer_zones(
    volume_element: dict,
    buffer_sizes: list[int],
    phase_ids: list[int],
    phase_labels: list[str],
    homog_label: str,
    order: list[str],
) -> dict:
    """Add buffer material regions to a volume element.

    Parameters
    ----------
    volume_element : dict
        Dict representing the volume element that can be validated via
        `validate_volume_element`.
    buffer_sizes : list of int, length 6
        Size of buffer on each face [-x, +x, -y, +y, -z, +z]
    phase_ids : list of int, length 6
        Phase of each buffer. Relative, so 1 is the first new phase and so on
    phase_labels : list of str
        Labels of the new phases
    homog_label: str
        Homogenization scheme label.
    order : list of str
        Order to add the zones in.

    Returns
    -------
    volume_element : dict
        Dict representing modified volume element.

    """
    volume_element = add_volume_element_buffer_zones(
        volume_element=volume_element,
        buffer_sizes=buffer_sizes,
        phase_ids=phase_ids,
        phase_labels=phase_labels,
        homog_label=homog_label,
        order=order,
    )
    volume_element["grid_size"] = tuple(int(i) for i in volume_element["grid_size"])
    volume_element["size"] = tuple(float(i) for i in volume_element["size"])
    return {"volume_element": volume_element}



import pprint

import numpy as np


def increment_chain_inner(x, g, all_x, all_g, all_x_inner, all_g_inner, threshold):

    if all_x_inner is None:
        # first "inner" iteration, need to set initial value:
        all_x_inner = np.array(all_x[-1])[None]
        all_g_inner = np.array([all_g[-1]])

    current_x = all_x_inner[-1]
    current_g = all_g_inner[-1]

    if g is None:
        # failed system analysis, reject the state:
        is_accept = False
        print(f"Increment chain inner: failed system analysis, rejecting state.")
    else:
        trial_x = x[:]  # convert to numpy array
        trial_g = g
        is_accept = trial_g > threshold

    new_x = trial_x if is_accept else current_x
    new_g = trial_g if is_accept else current_g

    all_x_inner = np.vstack([all_x_inner, new_x[None]])
    all_g_inner = np.append(all_g_inner, np.array(new_g))

    return {
        "x": new_x,
        "all_x_inner": all_x_inner,
        "all_g_inner": all_g_inner,
    }



import pytest
import matflow as mf


@pytest.mark.integration
def test_subset_sim_toy_model_two_level_external_parameter_flow(
    new_null_config, tmp_path
):

    # use `new_null_config` because currently these workflows include template components
    wk = mf.make_demo_workflow(
        "subset_simulation_toy_model_two_level_external", path=tmp_path
    )

    num_elements = wk.tasks.initialise_markov_chains.num_elements

    for elem_idx in range(num_elements):

        # input of `generate_next_state` from either `initialise_markov_chains`,
        # `increment_chain_inner`, or `increment_chain`:

        init_iters = {
            i.loop_idx["levels"]: i
            for i in wk.tasks.initialise_markov_chains.elements[elem_idx].iterations
        }
        gen_next_iters = {
            (
                i.loop_idx["inner_markov_chain"],
                i.loop_idx["outer_markov_chain"],
                i.loop_idx["levels"],
            ): i
            for i in wk.tasks.generate_next_state.elements[elem_idx].iterations
        }
        inc_chain_inner_iters = {
            (
                i.loop_idx["inner_markov_chain"],
                i.loop_idx["outer_markov_chain"],
                i.loop_idx["levels"],
            ): i
            for i in wk.tasks.increment_chain_inner.elements[elem_idx].iterations
        }
        inc_chain_iters = {
            (i.loop_idx["outer_markov_chain"], i.loop_idx["levels"]): i
            for i in wk.tasks.increment_chain.elements[elem_idx].iterations
        }

        for l_idx, iter_i in gen_next_iters.items():
            if (
                l_idx[0] == 0 and l_idx[1] == 0
            ):  # inner and outer MC loop zeroth iterations
                assert (
                    iter_i.get_data_idx()["inputs.x"]
                    == init_iters[l_idx[2]].get_data_idx()["outputs.x"]
                )

            elif l_idx[0] == 0:  # inner MC loop zeroth iteration, outer MC loop non-zero
                # should be sourced from `increment_chain`
                for lj_idx, iter_j in inc_chain_iters.items():
                    if lj_idx[1] == l_idx[2] and lj_idx[0] == l_idx[1] - 1:
                        assert (
                            iter_i.get_data_idx()["inputs.x"]
                            == iter_j.get_data_idx()["outputs.x"]
                        )

            elif l_idx[0] > 0:  # inner MC loop non-zero iteration:
                # should be sourced from `increment_chain_inner`, previous inner iteration, same outer MC loop iteration
                for lj_idx, iter_j in inc_chain_inner_iters.items():
                    if (
                        lj_idx[2] == l_idx[2]
                        and lj_idx[0] == l_idx[0] - 1
                        and lj_idx[1] == l_idx[1]
                    ):
                        assert (
                            iter_i.get_data_idx()["inputs.x"]
                            == iter_j.get_data_idx()["outputs.x"]
                        )
            else:
                raise RuntimeError(f"no test for loop index: {l_idx}")

        # new state from generate_next_state inner MC used in inner MC system_analysis and increment_chain_inner
        # keys are: (inner_markov_chain, outer_markov_chain, levels)
        gen_next_iters = {
            (
                i.loop_idx["inner_markov_chain"],
                i.loop_idx["outer_markov_chain"],
                i.loop_idx["levels"],
            ): i
            for i in wk.tasks.generate_next_state.elements[elem_idx].iterations
        }
        pre_proc_iters = {
            (
                i.loop_idx["inner_markov_chain"],
                i.loop_idx["outer_markov_chain"],
                i.loop_idx["levels"],
            ): i
            for i in wk.tasks.dummy_pre_processor_2.elements[elem_idx].iterations
        }
        model_iters = {
            (
                i.loop_idx["inner_markov_chain"],
                i.loop_idx["outer_markov_chain"],
                i.loop_idx["levels"],
            ): i
            for i in wk.tasks.model_sum_x_2.elements[elem_idx].iterations
        }
        limit_iters = {
            (
                i.loop_idx["inner_markov_chain"],
                i.loop_idx["outer_markov_chain"],
                i.loop_idx["levels"],
            ): i
            for i in wk.tasks.evaluate_limit_state_normal_sum_2.elements[
                elem_idx
            ].iterations
        }

        # check parameter flow within inner MC chain:

        for l_idx, iter_i in gen_next_iters.items():
            assert (
                pre_proc_iters[l_idx].get_data_idx()["inputs.x"]
                == iter_i.get_data_idx()["outputs.x"]
            )
            assert (
                inc_chain_inner_iters[l_idx].get_data_idx()["inputs.x"]
                == iter_i.get_data_idx()["outputs.x"]
            )

        for l_idx, iter_i in pre_proc_iters.items():
            assert (
                model_iters[l_idx].get_data_idx()["inputs.y"]
                == iter_i.get_data_idx()["outputs.y"]
            )

        for l_idx, iter_i in model_iters.items():
            assert (
                limit_iters[l_idx].get_data_idx()["inputs.sum_y"]
                == iter_i.get_data_idx()["outputs.sum_y"]
            )

        for l_idx, iter_i in limit_iters.items():
            assert (
                inc_chain_inner_iters[l_idx].get_data_idx()["inputs.g"]
                == iter_i.get_data_idx()["outputs.g"]
            )

        # check parameter flow into outer MC chain from inner MC chain:
        # keys are: (outer_markov_chain, levels)

        pre_proc_iters = {
            (i.loop_idx["outer_markov_chain"], i.loop_idx["levels"]): i
            for i in wk.tasks.dummy_pre_processor_3.elements[elem_idx].iterations
        }
        model_iters = {
            (i.loop_idx["outer_markov_chain"], i.loop_idx["levels"]): i
            for i in wk.tasks.model_sum_x_3.elements[elem_idx].iterations
        }
        limit_iters = {
            (i.loop_idx["outer_markov_chain"], i.loop_idx["levels"]): i
            for i in wk.tasks.evaluate_limit_state_normal_sum_3.elements[
                elem_idx
            ].iterations
        }

        # final inner MC chain iterations only:
        inc_chain_inner_max_iters = {}
        max_inner_iter = -1
        for i in wk.tasks.increment_chain_inner.elements[elem_idx].iterations:
            if i.loop_idx["inner_markov_chain"] > max_inner_iter:
                inc_chain_inner_max_iters[
                    (i.loop_idx["outer_markov_chain"], i.loop_idx["levels"])
                ] = i

        for l_idx, iter_i in inc_chain_inner_max_iters.items():
            assert (
                pre_proc_iters[l_idx].get_data_idx()["inputs.x"]
                == iter_i.get_data_idx()["outputs.x"]
            )

        for l_idx, iter_i in pre_proc_iters.items():
            assert (
                model_iters[l_idx].get_data_idx()["inputs.y"]
                == iter_i.get_data_idx()["outputs.y"]
            )

        for l_idx, iter_i in model_iters.items():
            assert (
                limit_iters[l_idx].get_data_idx()["inputs.sum_y"]
                == iter_i.get_data_idx()["outputs.sum_y"]
            )

        for l_idx, iter_i in limit_iters.items():
            assert (
                inc_chain_iters[l_idx].get_data_idx()["inputs.g"]
                == iter_i.get_data_idx()["outputs.g"]
            )


@pytest.mark.integration
def test_subset_sim_DAMASK_Mg_two_level_parameter_flow(new_null_config, tmp_path):

    # use `new_null_config` because currently these workflows include template components
    wk = mf.make_demo_workflow("subset_simulation_DAMASK_Mg_two_level", path=tmp_path)

    num_elements = wk.tasks.initialise_markov_chains.num_elements

    for elem_idx in range(num_elements):

        # input of `generate_next_state` from either `initialise_markov_chains`,
        # `increment_chain_inner`, or `increment_chain`:

        init_iters = {
            i.loop_idx["levels"]: i
            for i in wk.tasks.initialise_markov_chains.elements[elem_idx].iterations
        }
        gen_next_iters = {
            (
                i.loop_idx["inner_markov_chain"],
                i.loop_idx["outer_markov_chain"],
                i.loop_idx["levels"],
            ): i
            for i in wk.tasks.generate_next_state.elements[elem_idx].iterations
        }
        inc_chain_inner_iters = {
            (
                i.loop_idx["inner_markov_chain"],
                i.loop_idx["outer_markov_chain"],
                i.loop_idx["levels"],
            ): i
            for i in wk.tasks.increment_chain_inner.elements[elem_idx].iterations
        }
        inc_chain_iters = {
            (i.loop_idx["outer_markov_chain"], i.loop_idx["levels"]): i
            for i in wk.tasks.increment_chain.elements[elem_idx].iterations
        }

        for l_idx, iter_i in gen_next_iters.items():
            if (
                l_idx[0] == 0 and l_idx[1] == 0
            ):  # inner and outer MC loop zeroth iterations
                assert (
                    iter_i.get_data_idx()["inputs.x"]
                    == init_iters[l_idx[2]].get_data_idx()["outputs.x"]
                )

            elif l_idx[0] == 0:  # inner MC loop zeroth iteration, outer MC loop non-zero
                # should be sourced from `increment_chain`
                for lj_idx, iter_j in inc_chain_iters.items():
                    if lj_idx[1] == l_idx[2] and lj_idx[0] == l_idx[1] - 1:
                        assert (
                            iter_i.get_data_idx()["inputs.x"]
                            == iter_j.get_data_idx()["outputs.x"]
                        )

            elif l_idx[0] > 0:  # inner MC loop non-zero iteration:
                # should be sourced from `increment_chain_inner`, previous inner iteration, same outer MC loop iteration
                for lj_idx, iter_j in inc_chain_inner_iters.items():
                    if (
                        lj_idx[2] == l_idx[2]
                        and lj_idx[0] == l_idx[0] - 1
                        and lj_idx[1] == l_idx[1]
                    ):
                        assert (
                            iter_i.get_data_idx()["inputs.x"]
                            == iter_j.get_data_idx()["outputs.x"]
                        )
            else:
                raise RuntimeError(f"no test for loop index: {l_idx}")

        # new state from generate_next_state inner MC used in inner MC system_analysis and increment_chain_inner
        # keys are: (inner_markov_chain, outer_markov_chain, levels)
        gen_next_iters = {
            (
                i.loop_idx["inner_markov_chain"],
                i.loop_idx["outer_markov_chain"],
                i.loop_idx["levels"],
            ): i
            for i in wk.tasks.generate_next_state.elements[elem_idx].iterations
        }
        pre_proc_iters = {
            (
                i.loop_idx["inner_markov_chain"],
                i.loop_idx["outer_markov_chain"],
                i.loop_idx["levels"],
            ): i
            for i in wk.tasks.generate_volume_element_from_voronoi_random_variates_2.elements[
                elem_idx
            ].iterations
        }
        model_iters = {
            (
                i.loop_idx["inner_markov_chain"],
                i.loop_idx["outer_markov_chain"],
                i.loop_idx["levels"],
            ): i
            for i in wk.tasks.simulate_VE_loading_damask_2.elements[elem_idx].iterations
        }
        limit_iters = {
            (
                i.loop_idx["inner_markov_chain"],
                i.loop_idx["outer_markov_chain"],
                i.loop_idx["levels"],
            ): i
            for i in wk.tasks.evaluate_yield_stress_limit_state_2.elements[
                elem_idx
            ].iterations
        }

        # check parameter flow within inner MC chain:

        for l_idx, iter_i in gen_next_iters.items():
            assert (
                pre_proc_iters[l_idx].get_data_idx()["inputs.x"]
                == iter_i.get_data_idx()["outputs.x"]
            )
            assert (
                inc_chain_inner_iters[l_idx].get_data_idx()["inputs.x"]
                == iter_i.get_data_idx()["outputs.x"]
            )

        for l_idx, iter_i in pre_proc_iters.items():
            assert (
                model_iters[l_idx].get_data_idx()["inputs.volume_element"]
                == iter_i.get_data_idx()["outputs.volume_element"]
            )

        for l_idx, iter_i in model_iters.items():
            assert (
                limit_iters[l_idx].get_data_idx()["inputs.yield_stress"]
                == iter_i.get_data_idx()["outputs.yield_stress"]
            )

        for l_idx, iter_i in limit_iters.items():
            assert (
                inc_chain_inner_iters[l_idx].get_data_idx()["inputs.g"]
                == iter_i.get_data_idx()["outputs.g"]
            )

        # check parameter flow into outer MC chain from inner MC chain:
        # keys are: (outer_markov_chain, levels)

        pre_proc_iters = {
            (i.loop_idx["outer_markov_chain"], i.loop_idx["levels"]): i
            for i in wk.tasks.generate_volume_element_from_voronoi_random_variates_3.elements[
                elem_idx
            ].iterations
        }
        model_iters = {
            (i.loop_idx["outer_markov_chain"], i.loop_idx["levels"]): i
            for i in wk.tasks.simulate_VE_loading_damask_3.elements[elem_idx].iterations
        }
        limit_iters = {
            (i.loop_idx["outer_markov_chain"], i.loop_idx["levels"]): i
            for i in wk.tasks.evaluate_yield_stress_limit_state_3.elements[
                elem_idx
            ].iterations
        }

        # final inner MC chain iterations only:
        inc_chain_inner_max_iters = {}
        max_inner_iter = wk.loops.inner_markov_chain.num_iterations - 1
        for i in wk.tasks.increment_chain_inner.elements[elem_idx].iterations:
            if i.loop_idx["inner_markov_chain"] == max_inner_iter:
                inc_chain_inner_max_iters[
                    (i.loop_idx["outer_markov_chain"], i.loop_idx["levels"])
                ] = i

        for l_idx, iter_i in inc_chain_inner_max_iters.items():
            assert (
                pre_proc_iters[l_idx].get_data_idx()["inputs.x"]
                == iter_i.get_data_idx()["outputs.x"]
            )
            assert (
                inc_chain_iters[l_idx].get_data_idx()["inputs.x"]
                == iter_i.get_data_idx()["outputs.x"]
            )

        for l_idx, iter_i in pre_proc_iters.items():
            assert (
                model_iters[l_idx].get_data_idx()["inputs.volume_element"]
                == iter_i.get_data_idx()["outputs.volume_element"]
            )

        for l_idx, iter_i in model_iters.items():
            assert (
                limit_iters[l_idx].get_data_idx()["inputs.yield_stress"]
                == iter_i.get_data_idx()["outputs.yield_stress"]
            )

        for l_idx, iter_i in limit_iters.items():
            assert (
                inc_chain_iters[l_idx].get_data_idx()["inputs.g"]
                == iter_i.get_data_idx()["outputs.g"]
            )



"""
Configuration and standard fixtures for PyTest.
"""

from __future__ import annotations
from pathlib import Path
import numpy as np
import pytest
import matflow as mf
from matflow.param_classes.load import LoadCase, LoadStep
from matflow.param_classes.orientations import (
    EulerDefinition,
    LatticeDirection,
    OrientationRepresentation,
    OrientationRepresentationType,
    Orientations,
    UnitCellAlignment,
    QuatOrder,
)
from matflow.param_classes.seeds import MicrostructureSeeds


def pytest_addoption(parser: pytest.Parser):
    parser.addoption(
        "--integration",
        action="store_true",
        default=False,
        help="run integration-like workflow submission tests",
    )


def pytest_configure(config: pytest.Config):
    config.addinivalue_line(
        "markers",
        "integration: mark test as an integration-like workflow submission test to run",
    )
    mf.run_time_info.in_pytest = True


def pytest_collection_modifyitems(config: pytest.Config, items: list[pytest.Item]):
    if config.getoption("--integration"):
        # --integration in CLI: only run these tests
        for item in items:
            if "integration" not in item.keywords:
                item.add_marker(
                    pytest.mark.skip(reason="remove --integration option to run")
                )
    else:
        for item in items:
            if "integration" in item.keywords:
                item.add_marker(
                    pytest.mark.skip(reason="add --integration option to run")
                )


def pytest_unconfigure(config: pytest.Config):
    mf.run_time_info.in_pytest = False


@pytest.fixture
def null_config(tmp_path: Path):
    if not mf.is_config_loaded:
        mf.load_config(config_dir=tmp_path)
    mf.run_time_info.in_pytest = True


@pytest.fixture
def new_null_config(tmp_path: Path):
    mf.load_config(config_dir=tmp_path)
    mf.load_template_components(warn=False)
    mf.run_time_info.in_pytest = True


@pytest.fixture
def load_case_1() -> LoadCase:
    """A load case object to compare to that generated in `define_load.yaml`."""
    return LoadCase(
        steps=[
            LoadStep(
                total_time=100,
                num_increments=200,
                target_def_grad_rate=np.ma.masked_array(
                    data=np.array(
                        [
                            [1e-3, 0, 0],
                            [0, 0, 0],
                            [0, 0, 0],
                        ]
                    ),
                    mask=np.array(
                        [
                            [False, False, False],
                            [False, True, False],
                            [False, False, True],
                        ]
                    ),
                ),
                stress=np.ma.masked_array(
                    data=np.array(
                        [
                            [0, 0, 0],
                            [0, 0.0, 0],
                            [0, 0, 0.0],
                        ]
                    ),
                    mask=np.array(
                        [
                            [True, True, True],
                            [True, False, True],
                            [True, True, False],
                        ]
                    ),
                ),
            )
        ]
    )


@pytest.fixture
def orientations_1() -> Orientations:
    """An orientations object to compare to that generated in task index 0 of
    `define_orientations.yaml`."""
    return Orientations(
        data=np.array(
            [
                [0, 0, 0],
                [0, 45, 0],
            ]
        ),
        unit_cell_alignment=UnitCellAlignment(x=LatticeDirection.A),
        representation=OrientationRepresentation(
            type=OrientationRepresentationType.EULER,
            euler_definition=EulerDefinition.BUNGE,
            euler_is_degrees=True,
        ),
    )


@pytest.fixture
def orientations_2() -> Orientations:
    """An orientations object to compare to that generated in task index 1 of
    `define_orientations.yaml` (the demo data file `quaternions.txt`)."""
    return Orientations(
        data=np.array(
            [
                [
                    0.979576633518360,
                    -0.011699484277401,
                    -0.031022749430343,
                    0.198318758946959,
                ],
                [
                    0.051741844582538,
                    0.964477514397002,
                    0.258166574789950,
                    0.021352409770402,
                ],
                [
                    0.051741844582538,
                    0.964477514397002,
                    0.258166574789950,
                    0.021352409770402,
                ],
            ]
        ),
        unit_cell_alignment=UnitCellAlignment(
            x=LatticeDirection.A,
            y=LatticeDirection.B,
            z=LatticeDirection.C,
        ),
        representation=OrientationRepresentation(
            type=OrientationRepresentationType.QUATERNION,
            quat_order=QuatOrder.VECTOR_SCALAR,
        ),
    )


@pytest.fixture
def seeds_1(orientations_1: Orientations) -> MicrostructureSeeds:
    """A microstructure seeds object to compare to that generated in `define_seeds.yaml`."""
    return MicrostructureSeeds(
        position=np.array(
            [
                [0.3, 0.2, 0.1],
                [0.5, 0.4, 0.3],
            ]
        ),
        box_size=np.array([1.0, 1.0, 1.0]),
        phase_label="phase_1",
        orientations=orientations_1,
    )
